package jalview.datamodel;

import jalview.util.Comparison;
import jalview.util.ShiftList;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Vector;

public class HiddenColumns
{
  /*
   * list of hidden column [start, end] ranges; the list is maintained in
   * ascending start column order
   */
  private Vector<int[]> hiddenColumns;

  /**
   * This Method is used to return all the HiddenColumn regions
   * 
   * @return empty list or List of hidden column intervals
   */
  public List<int[]> getListOfCols()
  {
    return hiddenColumns == null ? Collections.<int[]> emptyList()
            : hiddenColumns;
  }

  public boolean hasHidden()
  {
    return (hiddenColumns != null) && (!hiddenColumns.isEmpty());
  }

  @Override
  public boolean equals(Object obj)
  {
    if (!(obj instanceof HiddenColumns))
    {
      return false;
    }
    HiddenColumns that = (HiddenColumns) obj;

    /*
     * check hidden columns are either both null, or match
     */
    if (this.hiddenColumns == null)
    {
      return (that.hiddenColumns == null);
    }
    if (that.hiddenColumns == null
            || that.hiddenColumns.size() != this.hiddenColumns.size())
    {
      return false;
    }
    int i = 0;
    for (int[] thisRange : hiddenColumns)
    {
      int[] thatRange = that.hiddenColumns.get(i++);
      if (thisRange[0] != thatRange[0] || thisRange[1] != thatRange[1])
      {
        return false;
      }
    }
    return true;
  }

  /**
   * Return absolute column index for a visible column index
   * 
   * @param column
   *          int column index in alignment view (count from zero)
   * @return alignment column index for column
   */
  public int adjustForHiddenColumns(int column)
  {
    int result = column;
    if (hiddenColumns != null)
    {
      for (int i = 0; i < hiddenColumns.size(); i++)
      {
        int[] region = hiddenColumns.elementAt(i);
        if (result >= region[0])
        {
          result += region[1] - region[0] + 1;
        }
      }
    }
    return result;
  }

  /**
   * Use this method to find out where a column will appear in the visible
   * alignment when hidden columns exist. If the column is not visible, then the
   * left-most visible column will always be returned.
   * 
   * @param hiddenColumn
   *          the column index in the full alignment including hidden columns
   * @return the position of the column in the visible alignment
   */
  public int findColumnPosition(int hiddenColumn)
  {
    int result = hiddenColumn;
    if (hiddenColumns != null)
    {
      int index = 0;
      int[] region;
      do
      {
        region = hiddenColumns.elementAt(index++);
        if (hiddenColumn > region[1])
        {
          result -= region[1] + 1 - region[0];
        }
      } while ((hiddenColumn > region[1]) && (index < hiddenColumns.size()));

      if (hiddenColumn >= region[0] && hiddenColumn <= region[1])
      {
        // Here the hidden column is within a region, so
        // we want to return the position of region[0]-1, adjusted for any
        // earlier hidden columns.
        // Calculate the difference between the actual hidden col position
        // and region[0]-1, and then subtract from result to convert result from
        // the adjusted hiddenColumn value to the adjusted region[0]-1 value

        // However, if the region begins at 0 we cannot return region[0]-1
        // just return 0
        if (region[0] == 0)
        {
          return 0;
        }
        else
        {
          return result - (hiddenColumn - region[0] + 1);
        }
      }
    }
    return result; // return the shifted position after removing hidden columns.
  }

  /**
   * Find the visible column which is a given visible number of columns to the
   * left of another visible column. i.e. for a startColumn x, the column which
   * is distance 1 away will be column x-1.
   * 
   * @param visibleDistance
   *          the number of visible columns to offset by
   * @param startColumn
   *          the column to start from
   * @return the position of the column in the visible alignment
   */
  public int subtractVisibleColumns(int visibleDistance, int startColumn)
  {
    int distance = visibleDistance;

    // in case startColumn is in a hidden region, move it to the left
    int start = adjustForHiddenColumns(findColumnPosition(startColumn));

    // get index of hidden region to left of start
    int index = getHiddenIndexLeft(start);
    if (index == -1)
    {
      // no hidden regions to left of startColumn
      return start - distance;
    }

    // walk backwards through the alignment subtracting the counts of visible
    // columns from distance
    int[] region;
    int gap = 0;
    int nextstart = start;

    while ((index > -1) && (distance - gap > 0))
    {
      // subtract the gap to right of region from distance
      distance -= gap;
      start = nextstart;

      // calculate the next gap
      region = hiddenColumns.get(index);
      gap = start - region[1];

      // set start to just to left of current region
      nextstart = region[0] - 1;
      index--;
    }

    if (distance - gap > 0)
    {
      // fell out of loop because there are no more hidden regions
      distance -= gap;
      return nextstart - distance;
    }
    return start - distance;

  }

  /**
   * Use this method to determine where the next hiddenRegion starts
   * 
   * @param hiddenRegion
   *          index of hidden region (counts from 0)
   * @return column number in visible view
   */
  public int findHiddenRegionPosition(int hiddenRegion)
  {
    int result = 0;
    if (hiddenColumns != null)
    {
      int index = 0;
      int gaps = 0;
      do
      {
        int[] region = hiddenColumns.elementAt(index);
        if (hiddenRegion == 0)
        {
          return region[0];
        }

        gaps += region[1] + 1 - region[0];
        result = region[1] + 1;
        index++;
      } while (index <= hiddenRegion);

      result -= gaps;
    }

    return result;
  }

  /**
   * This method returns the rightmost limit of a region of an alignment with
   * hidden columns. In otherwords, the next hidden column.
   * 
   * @param index
   *          int
   */
  public int getHiddenBoundaryRight(int alPos)
  {
    if (hiddenColumns != null)
    {
      int index = 0;
      do
      {
        int[] region = hiddenColumns.elementAt(index);
        if (alPos < region[0])
        {
          return region[0];
        }

        index++;
      } while (index < hiddenColumns.size());
    }

    return alPos;

  }

  /**
   * This method returns the leftmost limit of a region of an alignment with
   * hidden columns. In otherwords, the previous hidden column.
   * 
   * @param index
   *          int
   */
  public int getHiddenBoundaryLeft(int alPos)
  {
    if (hiddenColumns != null)
    {
      int index = hiddenColumns.size() - 1;
      do
      {
        int[] region = hiddenColumns.elementAt(index);
        if (alPos > region[1])
        {
          return region[1];
        }

        index--;
      } while (index > -1);
    }

    return alPos;

  }

  /**
   * This method returns the index of the hidden region to the left of a column
   * position. If the column is in a hidden region it returns the index of the
   * region to the left. If there is no hidden region to the left it returns -1.
   * 
   * @param pos
   *          int
   */
  private int getHiddenIndexLeft(int pos)
  {
    if (hiddenColumns != null)
    {
      int index = hiddenColumns.size() - 1;
      do
      {
        int[] region = hiddenColumns.elementAt(index);
        if (pos > region[1])
        {
          return index;
        }

        index--;
      } while (index > -1);
    }

    return -1;

  }

  /**
   * Adds the specified column range to the hidden columns
   * 
   * @param start
   * @param end
   */
  public void hideColumns(int start, int end)
  {
    if (hiddenColumns == null)
    {
      hiddenColumns = new Vector<int[]>();
    }

    /*
     * traverse existing hidden ranges and insert / amend / append as
     * appropriate
     */
    for (int i = 0; i < hiddenColumns.size(); i++)
    {
      int[] region = hiddenColumns.elementAt(i);

      if (end < region[0] - 1)
      {
        /*
         * insert discontiguous preceding range
         */
        hiddenColumns.insertElementAt(new int[] { start, end }, i);
        return;
      }

      if (end <= region[1])
      {
        /*
         * new range overlaps existing, or is contiguous preceding it - adjust
         * start column
         */
        region[0] = Math.min(region[0], start);
        return;
      }

      if (start <= region[1] + 1)
      {
        /*
         * new range overlaps existing, or is contiguous following it - adjust
         * start and end columns
         */
        region[0] = Math.min(region[0], start);
        region[1] = Math.max(region[1], end);

        /*
         * also update or remove any subsequent ranges 
         * that are overlapped
         */
        while (i < hiddenColumns.size() - 1)
        {
          int[] nextRegion = hiddenColumns.get(i + 1);
          if (nextRegion[0] > end + 1)
          {
            /*
             * gap to next hidden range - no more to update
             */
            break;
          }
          region[1] = Math.max(nextRegion[1], end);
          hiddenColumns.remove(i + 1);
        }
        return;
      }
    }

    /*
     * remaining case is that the new range follows everything else
     */
    hiddenColumns.addElement(new int[] { start, end });
  }

  public boolean isVisible(int column)
  {
    if (hiddenColumns != null)
    {
      for (int[] region : hiddenColumns)
      {
        if (column >= region[0] && column <= region[1])
        {
          return false;
        }
      }
    }

    return true;
  }

  /**
   * ColumnSelection
   */
  public HiddenColumns()
  {
  }

  /**
   * Copy constructor
   * 
   * @param copy
   */
  public HiddenColumns(HiddenColumns copy)
  {
    if (copy != null)
    {
      if (copy.hiddenColumns != null)
      {
        hiddenColumns = new Vector<int[]>(copy.hiddenColumns.size());
        for (int i = 0, j = copy.hiddenColumns.size(); i < j; i++)
        {
          int[] rh, cp;
          rh = copy.hiddenColumns.elementAt(i);
          if (rh != null)
          {
            cp = new int[rh.length];
            System.arraycopy(rh, 0, cp, 0, rh.length);
            hiddenColumns.addElement(cp);
          }
        }
      }
    }
  }

  /**
   * propagate shift in alignment columns to column selection
   * 
   * @param start
   *          beginning of edit
   * @param left
   *          shift in edit (+ve for removal, or -ve for inserts)
   */
  public List<int[]> compensateForEdit(int start, int change,
          ColumnSelection sel)
  {
    List<int[]> deletedHiddenColumns = null;

    if (hiddenColumns != null)
    {
      deletedHiddenColumns = new ArrayList<int[]>();
      int hSize = hiddenColumns.size();
      for (int i = 0; i < hSize; i++)
      {
        int[] region = hiddenColumns.elementAt(i);
        if (region[0] > start && start + change > region[1])
        {
          deletedHiddenColumns.add(region);

          hiddenColumns.removeElementAt(i);
          i--;
          hSize--;
          continue;
        }

        if (region[0] > start)
        {
          region[0] -= change;
          region[1] -= change;
        }

        if (region[0] < 0)
        {
          region[0] = 0;
        }

      }

      this.revealHiddenColumns(0, sel);
    }

    return deletedHiddenColumns;
  }

  /**
   * propagate shift in alignment columns to column selection special version of
   * compensateForEdit - allowing for edits within hidden regions
   * 
   * @param start
   *          beginning of edit
   * @param left
   *          shift in edit (+ve for removal, or -ve for inserts)
   */
  public void compensateForDelEdits(int start, int change)
  {
    if (hiddenColumns != null)
    {
      for (int i = 0; i < hiddenColumns.size(); i++)
      {
        int[] region = hiddenColumns.elementAt(i);
        if (region[0] >= start)
        {
          region[0] -= change;
        }
        if (region[1] >= start)
        {
          region[1] -= change;
        }
        if (region[1] < region[0])
        {
          hiddenColumns.removeElementAt(i--);
        }

        if (region[0] < 0)
        {
          region[0] = 0;
        }
        if (region[1] < 0)
        {
          region[1] = 0;
        }
      }
    }
  }

  /**
   * return all visible segments between the given start and end boundaries
   * 
   * @param start
   *          (first column inclusive from 0)
   * @param end
   *          (last column - not inclusive)
   * @return int[] {i_start, i_end, ..} where intervals lie in
   *         start<=i_start<=i_end<end
   */
  public int[] getVisibleContigs(int start, int end)
  {
    if (hiddenColumns != null && hiddenColumns.size() > 0)
    {
      List<int[]> visiblecontigs = new ArrayList<int[]>();
      List<int[]> regions = getListOfCols();

      int vstart = start;
      int[] region;
      int hideStart, hideEnd;

      for (int j = 0; vstart < end && j < regions.size(); j++)
      {
        region = regions.get(j);
        hideStart = region[0];
        hideEnd = region[1];

        if (hideEnd < vstart)
        {
          continue;
        }
        if (hideStart > vstart)
        {
          visiblecontigs.add(new int[] { vstart, hideStart - 1 });
        }
        vstart = hideEnd + 1;
      }

      if (vstart < end)
      {
        visiblecontigs.add(new int[] { vstart, end - 1 });
      }
      int[] vcontigs = new int[visiblecontigs.size() * 2];
      for (int i = 0, j = visiblecontigs.size(); i < j; i++)
      {
        int[] vc = visiblecontigs.get(i);
        visiblecontigs.set(i, null);
        vcontigs[i * 2] = vc[0];
        vcontigs[i * 2 + 1] = vc[1];
      }
      visiblecontigs.clear();
      return vcontigs;
    }
    else
    {
      return new int[] { start, end - 1 };
    }
  }

  public String[] getVisibleSequenceStrings(int start, int end,
          SequenceI[] seqs)
  {
    int i, iSize = seqs.length;
    String selections[] = new String[iSize];
    if (hiddenColumns != null && hiddenColumns.size() > 0)
    {
      for (i = 0; i < iSize; i++)
      {
        StringBuffer visibleSeq = new StringBuffer();
        List<int[]> regions = getListOfCols();

        int blockStart = start, blockEnd = end;
        int[] region;
        int hideStart, hideEnd;

        for (int j = 0; j < regions.size(); j++)
        {
          region = regions.get(j);
          hideStart = region[0];
          hideEnd = region[1];

          if (hideStart < start)
          {
            continue;
          }

          blockStart = Math.min(blockStart, hideEnd + 1);
          blockEnd = Math.min(blockEnd, hideStart);

          if (blockStart > blockEnd)
          {
            break;
          }

          visibleSeq.append(seqs[i].getSequence(blockStart, blockEnd));

          blockStart = hideEnd + 1;
          blockEnd = end;
        }

        if (end > blockStart)
        {
          visibleSeq.append(seqs[i].getSequence(blockStart, end));
        }

        selections[i] = visibleSeq.toString();
      }
    }
    else
    {
      for (i = 0; i < iSize; i++)
      {
        selections[i] = seqs[i].getSequenceAsString(start, end);
      }
    }

    return selections;
  }

  /**
   * Locate the first and last position visible for this sequence. if seq isn't
   * visible then return the position of the left and right of the hidden
   * boundary region, and the corresponding alignment column indices for the
   * extent of the sequence
   * 
   * @param seq
   * @return int[] { visible start, visible end, first seqpos, last seqpos,
   *         alignment index for seq start, alignment index for seq end }
   */
  public int[] locateVisibleBoundsOfSequence(SequenceI seq)
  {
    int fpos = seq.getStart(), lpos = seq.getEnd();
    int start = 0;

    if (hiddenColumns == null || hiddenColumns.size() == 0)
    {
      int ifpos = seq.findIndex(fpos) - 1, ilpos = seq.findIndex(lpos) - 1;
      return new int[] { ifpos, ilpos, fpos, lpos, ifpos, ilpos };
    }

    // Simply walk along the sequence whilst watching for hidden column
    // boundaries
    List<int[]> regions = getListOfCols();
    int spos = fpos, lastvispos = -1, rcount = 0, hideStart = seq
            .getLength(), hideEnd = -1;
    int visPrev = 0, visNext = 0, firstP = -1, lastP = -1;
    boolean foundStart = false;
    for (int p = 0, pLen = seq.getLength(); spos <= seq.getEnd()
            && p < pLen; p++)
    {
      if (!Comparison.isGap(seq.getCharAt(p)))
      {
        // keep track of first/last column
        // containing sequence data regardless of visibility
        if (firstP == -1)
        {
          firstP = p;
        }
        lastP = p;
        // update hidden region start/end
        while (hideEnd < p && rcount < regions.size())
        {
          int[] region = regions.get(rcount++);
          visPrev = visNext;
          visNext += region[0] - visPrev;
          hideStart = region[0];
          hideEnd = region[1];
        }
        if (hideEnd < p)
        {
          hideStart = seq.getLength();
        }
        // update visible boundary for sequence
        if (p < hideStart)
        {
          if (!foundStart)
          {
            fpos = spos;
            start = p;
            foundStart = true;
          }
          lastvispos = p;
          lpos = spos;
        }
        // look for next sequence position
        spos++;
      }
    }
    if (foundStart)
    {
      return new int[] { findColumnPosition(start),
          findColumnPosition(lastvispos), fpos, lpos, firstP, lastP };
    }
    // otherwise, sequence was completely hidden
    return new int[] { visPrev, visNext, 0, 0, firstP, lastP };
  }

  /**
   * delete any columns in alignmentAnnotation that are hidden (including
   * sequence associated annotation).
   * 
   * @param alignmentAnnotation
   */
  public void makeVisibleAnnotation(AlignmentAnnotation alignmentAnnotation)
  {
    makeVisibleAnnotation(-1, -1, alignmentAnnotation);
  }

  /**
   * delete any columns in alignmentAnnotation that are hidden (including
   * sequence associated annotation).
   * 
   * @param start
   *          remove any annotation to the right of this column
   * @param end
   *          remove any annotation to the left of this column
   * @param alignmentAnnotation
   *          the annotation to operate on
   */
  public void makeVisibleAnnotation(int start, int end,
          AlignmentAnnotation alignmentAnnotation)
  {
    if (alignmentAnnotation.annotations == null)
    {
      return;
    }
    if (start == end && end == -1)
    {
      start = 0;
      end = alignmentAnnotation.annotations.length;
    }
    if (hiddenColumns != null && hiddenColumns.size() > 0)
    {
      // then mangle the alignmentAnnotation annotation array
      Vector<Annotation[]> annels = new Vector<Annotation[]>();
      Annotation[] els = null;
      List<int[]> regions = getListOfCols();
      int blockStart = start, blockEnd = end;
      int[] region;
      int hideStart, hideEnd, w = 0;

      for (int j = 0; j < regions.size(); j++)
      {
        region = regions.get(j);
        hideStart = region[0];
        hideEnd = region[1];

        if (hideStart < start)
        {
          continue;
        }

        blockStart = Math.min(blockStart, hideEnd + 1);
        blockEnd = Math.min(blockEnd, hideStart);

        if (blockStart > blockEnd)
        {
          break;
        }

        annels.addElement(els = new Annotation[blockEnd - blockStart]);
        System.arraycopy(alignmentAnnotation.annotations, blockStart, els,
                0, els.length);
        w += els.length;
        blockStart = hideEnd + 1;
        blockEnd = end;
      }

      if (end > blockStart)
      {
        annels.addElement(els = new Annotation[end - blockStart + 1]);
        if ((els.length + blockStart) <= alignmentAnnotation.annotations.length)
        {
          // copy just the visible segment of the annotation row
          System.arraycopy(alignmentAnnotation.annotations, blockStart,
                  els, 0, els.length);
        }
        else
        {
          // copy to the end of the annotation row
          System.arraycopy(alignmentAnnotation.annotations, blockStart,
                  els, 0,
                  (alignmentAnnotation.annotations.length - blockStart));
        }
        w += els.length;
      }
      if (w == 0)
      {
        return;
      }

      alignmentAnnotation.annotations = new Annotation[w];
      w = 0;

      for (Annotation[] chnk : annels)
      {
        System.arraycopy(chnk, 0, alignmentAnnotation.annotations, w,
                chnk.length);
        w += chnk.length;
      }
    }
    else
    {
      alignmentAnnotation.restrict(start, end);
    }
  }

  /**
   * 
   * @return true if there are columns hidden
   */
  public boolean hasHiddenColumns()
  {
    return hiddenColumns != null && hiddenColumns.size() > 0;
  }

  /**
   * 
   * @return true if there are more than one set of columns hidden
   */
  public boolean hasManyHiddenColumns()
  {
    return hiddenColumns != null && hiddenColumns.size() > 1;
  }

  /**
   * mark the columns corresponding to gap characters as hidden in the column
   * selection
   * 
   * @param sr
   */
  public void hideInsertionsFor(SequenceI sr)
  {
    List<int[]> inserts = sr.getInsertions();
    for (int[] r : inserts)
    {
      hideColumns(r[0], r[1]);
    }
  }

  /**
   * Unhides, and adds to the selection list, all hidden columns
   */
  public void revealAllHiddenColumns(ColumnSelection sel)
  {
    if (hiddenColumns != null)
    {
      for (int i = 0; i < hiddenColumns.size(); i++)
      {
        int[] region = hiddenColumns.elementAt(i);
        for (int j = region[0]; j < region[1] + 1; j++)
        {
          sel.addElement(j);
        }
      }
    }

    hiddenColumns = null;
  }

  /**
   * Reveals, and marks as selected, the hidden column range with the given
   * start column
   * 
   * @param start
   */
  public void revealHiddenColumns(int start, ColumnSelection sel)
  {
    for (int i = 0; i < hiddenColumns.size(); i++)
    {
      int[] region = hiddenColumns.elementAt(i);
      if (start == region[0])
      {
        for (int j = region[0]; j < region[1] + 1; j++)
        {
          sel.addElement(j);
        }

        hiddenColumns.removeElement(region);
        break;
      }
    }
    if (hiddenColumns.size() == 0)
    {
      hiddenColumns = null;
    }
  }

  /**
   * removes intersection of position,length ranges in deletions from the
   * start,end regions marked in intervals.
   * 
   * @param shifts
   * @param intervals
   * @return
   */
  private boolean pruneIntervalVector(final List<int[]> shifts,
          Vector<int[]> intervals)
  {
    boolean pruned = false;
    int i = 0, j = intervals.size() - 1, s = 0, t = shifts.size() - 1;
    int hr[] = intervals.elementAt(i);
    int sr[] = shifts.get(s);
    while (i <= j && s <= t)
    {
      boolean trailinghn = hr[1] >= sr[0];
      if (!trailinghn)
      {
        if (i < j)
        {
          hr = intervals.elementAt(++i);
        }
        else
        {
          i++;
        }
        continue;
      }
      int endshift = sr[0] + sr[1]; // deletion ranges - -ve means an insert
      if (endshift < hr[0] || endshift < sr[0])
      { // leadinghc disjoint or not a deletion
        if (s < t)
        {
          sr = shifts.get(++s);
        }
        else
        {
          s++;
        }
        continue;
      }
      boolean leadinghn = hr[0] >= sr[0];
      boolean leadinghc = hr[0] < endshift;
      boolean trailinghc = hr[1] < endshift;
      if (leadinghn)
      {
        if (trailinghc)
        { // deleted hidden region.
          intervals.removeElementAt(i);
          pruned = true;
          j--;
          if (i <= j)
          {
            hr = intervals.elementAt(i);
          }
          continue;
        }
        if (leadinghc)
        {
          hr[0] = endshift; // clip c terminal region
          leadinghn = !leadinghn;
          pruned = true;
        }
      }
      if (!leadinghn)
      {
        if (trailinghc)
        {
          if (trailinghn)
          {
            hr[1] = sr[0] - 1;
            pruned = true;
          }
        }
        else
        {
          // sr contained in hr
          if (s < t)
          {
            sr = shifts.get(++s);
          }
          else
          {
            s++;
          }
          continue;
        }
      }
    }
    return pruned; // true if any interval was removed or modified by
    // operations.
  }

  /**
   * remove any hiddenColumns or selected columns and shift remaining based on a
   * series of position, range deletions.
   * 
   * @param deletions
   */
  public void pruneDeletions(List<int[]> shifts)
  {
    // delete any intervals intersecting.
    if (hiddenColumns != null)
    {
      pruneIntervalVector(shifts, hiddenColumns);
      if (hiddenColumns != null && hiddenColumns.size() == 0)
      {
        hiddenColumns = null;
      }
    }
  }

  /**
   * Add gaps into the sequences aligned to profileseq under the given
   * AlignmentView
   * 
   * @param profileseq
   * @param al
   *          - alignment to have gaps inserted into it
   * @param input
   *          - alignment view where sequence corresponding to profileseq is
   *          first entry
   * @return new HiddenColumns for new alignment view, with insertions into
   *         profileseq marked as hidden.
   */
  public static HiddenColumns propagateInsertions(SequenceI profileseq,
          AlignmentI al, AlignmentView input)
  {
    int profsqpos = 0;

    char gc = al.getGapCharacter();
    Object[] alandhidden = input.getAlignmentAndHiddenColumns(gc);
    HiddenColumns nview = (HiddenColumns) alandhidden[1];
    SequenceI origseq = ((SequenceI[]) alandhidden[0])[profsqpos];
    nview.propagateInsertions(profileseq, al, origseq);
    return nview;
  }

  /**
   * 
   * @param profileseq
   *          - sequence in al which corresponds to origseq
   * @param al
   *          - alignment which is to have gaps inserted into it
   * @param origseq
   *          - sequence corresponding to profileseq which defines gap map for
   *          modifying al
   */
  private void propagateInsertions(SequenceI profileseq, AlignmentI al,
          SequenceI origseq)
  {
    char gc = al.getGapCharacter();
    // recover mapping between sequence's non-gap positions and positions
    // mapping to view.
    pruneDeletions(ShiftList.parseMap(origseq.gapMap()));
    int[] viscontigs = al.getHiddenColumns().getVisibleContigs(0,
            profileseq.getLength());
    int spos = 0;
    int offset = 0;

    // add profile to visible contigs
    for (int v = 0; v < viscontigs.length; v += 2)
    {
      if (viscontigs[v] > spos)
      {
        StringBuffer sb = new StringBuffer();
        for (int s = 0, ns = viscontigs[v] - spos; s < ns; s++)
        {
          sb.append(gc);
        }
        for (int s = 0, ns = al.getHeight(); s < ns; s++)
        {
          SequenceI sqobj = al.getSequenceAt(s);
          if (sqobj != profileseq)
          {
            String sq = al.getSequenceAt(s).getSequenceAsString();
            if (sq.length() <= spos + offset)
            {
              // pad sequence
              int diff = spos + offset - sq.length() - 1;
              if (diff > 0)
              {
                // pad gaps
                sq = sq + sb;
                while ((diff = spos + offset - sq.length() - 1) > 0)
                {
                  // sq = sq
                  // + ((diff >= sb.length()) ? sb.toString() : sb
                  // .substring(0, diff));
                  if (diff >= sb.length())
                  {
                    sq += sb.toString();
                  }
                  else
                  {
                    char[] buf = new char[diff];
                    sb.getChars(0, diff, buf, 0);
                    sq += buf.toString();
                  }
                }
              }
              sq += sb.toString();
            }
            else
            {
              al.getSequenceAt(s).setSequence(
                      sq.substring(0, spos + offset) + sb.toString()
                              + sq.substring(spos + offset));
            }
          }
        }
        // offset+=sb.length();
      }
      spos = viscontigs[v + 1] + 1;
    }
    if ((offset + spos) < profileseq.getLength())
    {
      // pad the final region with gaps.
      StringBuffer sb = new StringBuffer();
      for (int s = 0, ns = profileseq.getLength() - spos - offset; s < ns; s++)
      {
        sb.append(gc);
      }
      for (int s = 0, ns = al.getHeight(); s < ns; s++)
      {
        SequenceI sqobj = al.getSequenceAt(s);
        if (sqobj == profileseq)
        {
          continue;
        }
        String sq = sqobj.getSequenceAsString();
        // pad sequence
        int diff = origseq.getLength() - sq.length();
        while (diff > 0)
        {
          // sq = sq
          // + ((diff >= sb.length()) ? sb.toString() : sb
          // .substring(0, diff));
          if (diff >= sb.length())
          {
            sq += sb.toString();
          }
          else
          {
            char[] buf = new char[diff];
            sb.getChars(0, diff, buf, 0);
            sq += buf.toString();
          }
          diff = origseq.getLength() - sq.length();
        }
      }
    }
  }

  /**
   * remove any hiddenColumns or selected columns and shift remaining based on a
   * series of position, range deletions.
   * 
   * @param deletions
   */
  private void pruneDeletions(ShiftList deletions)
  {
    if (deletions != null)
    {
      final List<int[]> shifts = deletions.getShifts();
      if (shifts != null && shifts.size() > 0)
      {
        pruneDeletions(shifts);

        // and shift the rest.
        this.compensateForEdits(deletions);
      }
    }
  }

  /**
   * Adjust hidden column boundaries based on a series of column additions or
   * deletions in visible regions.
   * 
   * @param shiftrecord
   * @return
   */
  private ShiftList compensateForEdits(ShiftList shiftrecord)
  {
    if (shiftrecord != null)
    {
      final List<int[]> shifts = shiftrecord.getShifts();
      if (shifts != null && shifts.size() > 0)
      {
        int shifted = 0;
        for (int i = 0, j = shifts.size(); i < j; i++)
        {
          int[] sh = shifts.get(i);
          compensateForDelEdits(shifted + sh[0], sh[1]);
          shifted -= sh[1];
        }
      }
      return shiftrecord.getInverse();
    }
    return null;
  }

  /**
   * Returns a hashCode built from hidden column ranges
   */
  public int hashCode(int hc)
  {
    int hashCode = hc;
    if (hiddenColumns != null)
    {
      for (int[] hidden : hiddenColumns)
      {
        hashCode = 31 * hashCode + hidden[0];
        hashCode = 31 * hashCode + hidden[1];
      }
    }
    return hashCode;
  }

}
